/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Matt Eagleson (e.Messiah@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <gnome.h>
#include <pthread.h>
#include <errno.h>
#include <ctype.h>

#include "../uulib/uudeview.h"
#include "../uulib/fptools.h"

#include "pan.h"

int must_exit = FALSE;
int stop_all = FALSE;
int online = FALSE;
static pthread_cond_t online_c = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t online_m = PTHREAD_MUTEX_INITIALIZER;

// extern GCompareFunc articlelist_sort_func;

/**************************************************
* go on/off line code
***************************************************/

void connect_dialog_clicked_cb(GnomeDialog *dialog, int button, int fd[2])
{
	int i;
	
	switch (button) {
	case 0:					/* OK button */
		i = 1;
		break;
	default:			/* Cancel Button */
		i = 0;
		break;
	}
	write(fd[1], &i, sizeof(int));
	gtk_widget_destroy(GTK_WIDGET(dialog));
}

void connect_wait (void)
{              
	pthread_mutex_lock(&online_m);
	pthread_cond_wait(&online_c, &online_m);
	pthread_mutex_unlock(&online_m);
}

int go_online (void)
{
	if (online == FALSE) {
		if (connect_to_server() == 0) {
			online = TRUE;
			gui_set_status_default(_("Pan - Online"));
			gui_clean();
			log_add("Pan - Online");
			pthread_mutex_lock(&online_m);
			pthread_cond_broadcast(&online_c);
			pthread_mutex_unlock(&online_m);
			return TRUE;
		} else {
			return FALSE;
		}
	} else {
		return TRUE;
	}
}

int
go_online_ask (void)
{
	GtkWidget *connect_dialog;
	GtkWidget *label;
	static int fd[2];
	int i = 0;

	if (online == FALSE) {
		/* We setup a pipe for communication between the threads */
		if (pipe(fd) != 0) {
			g_warning("Couldn't create pipe");
			return FALSE;
		}
		/* Spawn the go_online? dialog, make it modal so we don't
		 * leave the network thread in limbo forever */
		pan_lock();
		connect_dialog = gnome_dialog_new (_("Go online?"), GNOME_STOCK_BUTTON_YES, GNOME_STOCK_BUTTON_NO, NULL);
		label = gtk_label_new (_("A connection to the server is needed.\nShall we? "));
		gtk_container_add (GTK_CONTAINER(GNOME_DIALOG(connect_dialog)->vbox), label);
	   	gtk_signal_connect (GTK_OBJECT(connect_dialog), "clicked", 
				    GTK_SIGNAL_FUNC(connect_dialog_clicked_cb), fd);
		gtk_window_set_modal (GTK_WINDOW(connect_dialog), TRUE);

		gui_popup_draw (connect_dialog);
/*		gtk_widget_show_all (connect_dialog); */
		pan_unlock();

		read(fd[0], &i, sizeof(int));
		if (i == 1) {
			return go_online();
		} else {
			return FALSE;
		}
	} else {
		return TRUE;
	}
}

void go_offline (void)
{
	if (online == TRUE) {
		put_server("QUIT\r\n");
		tcp_close();
		must_exit = FALSE;
		gui_set_status_default(_("Pan - Offline"));
		log_add("Pan - Offline");
		gui_clean();
		if (stop_all) {
			queue_purge ();
			stop_all = FALSE;
		}
		online = FALSE;
	}
}

int is_online(void)
{
	return online;
}

void network_error(const char *format, ...)
{
	va_list args;
	char *p = NULL;

	g_return_if_fail(format != NULL);

	va_start(args, format);
	p = g_strdup_vprintf(format, args);
	va_end(args);

	pan_lock();
	gnome_error_dialog(p);
	pan_unlock();
	g_free(p);
	go_offline();
}

int set_group(group_data * data)
{
	char *buffer;

	g_assert(data);

	if (put_server_va("GROUP %s\r\n", data->name) == -1) {
		return 0;
	}
	if ((buffer = get_server()) == NULL) {
		return 0;
	}
	if (atoi(buffer) != 211)
		return 1;
	else
		return 0;
}


/***************************************************************
* article_post
****************************************************************/

int nntp_post(char *msg)
{
	char *buffer;

	gui_set_status(_("Posting message..."));
	if (put_server("POST\r\n") == -1) {
		network_error(_("Error writing to socket during post."));
		return -1;
	}
	if ( (buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket during post."));
		return -1;
	}
	if (atoi(buffer) == 340) {
		if (put_server_va("%s\r\n.\r\n", msg) == -1) {
			network_error(_("Error writing to socket during post."));
			return -1;
		}
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket during post."));
			return -1;
		}
		if (atoi(buffer) == 240) {
			return 0;
		}
	}
	gui_clean();
	pan_error_dialog(_("Error posting message.  Server said:\n'%s'"), buffer);
	return -1;
}


/***************************************************************
* grouplist_download 
****************************************************************/

int nntp_grouplist_download(void)
{
	char *buffer = NULL;
	group_data *gdata = NULL;
	int count = 0;
	GSList *glist = NULL;

	gui_set_status(_("Retrieved 0 groups."));

	if (put_server("LIST active\r\n") == -1) {
		network_error(_("Error writing to socket."));
		return -1;
	}
	if ( (buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return -1;
	}
	if (atoi(buffer) != 215) {
		network_error(_("Group list could not be retrieved\nServer said: %s"), buffer);
		return -1;
	} else {
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket."));
			return -1;
		}
		gui_set_progress_activity(1);
		while (strncmp(buffer, ".\r\n", 3)) {
			if (must_exit) {
				grouplist_free(glist);
				go_offline();
				return -1;
			}
			gdata = (group_data *) g_malloc0(sizeof(group_data));
			gdata->name = g_strdup(strtok(buffer, " "));
			glist = g_slist_prepend(glist, gdata);
			gui_set_progress_activity(++count);
			if ((count % 23) == 0) {
				gui_set_status_va(_("Retrieved %d groups."), count);
			}
			if ((buffer = get_server()) == NULL) {
				network_error(_("Error reading from socket."));
				grouplist_free(glist);
				return -1;
			}
		}
	}
	gui_set_info_va(_("Retrieved %d groups."), count);
	log_add_va("Retrieved %d groups", count);

	Pan.grouplist = glist;
	return count;
}


/***************************************************************
* articlelist_download 
****************************************************************/

int nntp_articlelist_download(group_data *gdata, int flags)
{
	char *buffer = NULL;
	article_data *adata = NULL;
	int count = 0, total;
	int first = 0;
	int last = 0;
	int p_last = 0;
	int part = 0;
	int parts = 0;
	int i;
	time_t current_time, date;
	char **sd;
	char *s;
	int expiration;
	GSList *alist = NULL;

	expiration = gnome_config_get_int_with_default ("/Pan/Articles/Expiration=30", NULL);

	if (put_server_va("GROUP %s\r\n", gdata->name) == -1) {
		network_error(_("Error writing to socket."));
		return -1;
	}
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return -1;
	}

	if (strtol(buffer, NULL, 10) != 211) {
		network_error(_("Couldn't select group '%s'"), gdata->name);
		return -1;
	}

	sscanf(buffer, "%*d %d %d %d", &total, &first, &last);

	p_last = group_get_attrib_i (gdata, "Last");
	if (total == '0') {
		gui_set_info(_("No articles"));
		return -1;
	}
	if ((flags & QUEUE_ARTICLELIST_DOWNLOAD_NEW) && p_last) {
		if (p_last == last) {
			gui_set_info(_("No new articles"));
			return -1;
		}
		gui_set_status(_("Retrieving New Article Headers"));
		if (put_server_va("XOVER %d-%d\r\n", p_last, last) == -1) {
			network_error(_("Error writing to socket."));
			return -1;
		}
	} else {
		if ((total > 150) && ((flags & 0x003F) == QUEUE_ARTICLELIST_DOWNLOAD_SAMPLE)) {
			gui_set_status(_("Retrieving Sample of Article Headers"));
			if (put_server_va("XOVER %d-%d\r\n", first, first + 150) == -1) {
				network_error(_("Error writing to socket."));
				return -1;
			}
		} else {
			gui_set_status(_("Retrieving All Article Headers"));
			if (put_server_va("XOVER %d-%d\r\n", first, last) == -1) {
				network_error(_("Error writing to socket."));
				return -1;
			}
		}
	}

	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return -1;
	}

	if (strtol(buffer, (char **) NULL, 10) != 224) {
		network_error(_("Couldn't retrieve headers\nServer said: %s"), buffer);
		return -1;
	}

	time(&current_time);
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return -1;
	}

	while (strncmp(buffer, ".\r\n", 3)) {
		if (must_exit) {
			articlelist_free(alist);
			go_offline();
			return -1;
		}
		sd = g_strsplit(buffer, "\t", 0);
		date = parse_date(sd[3]);
		if ((current_time - date) > (86400 * expiration)) {
			g_strfreev(sd);
			if ((buffer = get_server()) == NULL) {
				network_error(_("Error reading from socket."));
				articlelist_free(alist);
				g_strfreev(sd);
				return -1;
			}
			continue;
		}
		adata = (article_data *) g_malloc0(sizeof(article_data));
		adata->number = g_strdup(sd[0]);
		adata->subject = g_strdup(sd[1]);
		adata->author = g_strdup(sd[2]);
		adata->date = date;
		adata->message_id = g_strdup(sd[4]);
		adata->references = g_strdup(sd[5]);
		adata->linecount = strtol(sd[7], NULL, 10);
		adata->group = gdata;
		adata->state = STATE_NEW;

		g_strfreev(sd);

		/* Here we look for the (n/N) or [n/N] construct in subject lines,
		 * starting at the end of the string and working backwards */
		part = 0;
		parts = 0;
		s = adata->subject;
		for (i = strlen(s) - 1; i > 0; --i) {
			if ((s[i] == ')') || (s[i] == ']')) {
				for (--i; isdigit(s[i]); --i);
				if (s[i] == '/') {
					parts = atoi(s + i + 1);
					for (--i; isdigit(s[i]); --i);
					if ((s[i] == '(') || (s[i] == '[')) {
						part = atoi(s + i + 1);
						break;
					}
				}
			}
		}

		/* Verify Multipart info */
		if ((parts > 1) && (part <= parts)) {
			/* We found one! */
			adata->parts = parts;
			adata->part = part;
		} else {
			adata->parts = 1;
			adata->part = 1;
		}
		/* Add the article to the article list */
		alist = g_slist_prepend(alist, adata);

		gui_set_progress((float)++ count / (float) total);
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket."));
			articlelist_free(alist);
			return -1;
		}
	}
	/* Check to see if all articles were expired */
	if (alist == NULL) {
		gui_set_info(_("All articles were expired"));
		return -1;
	}
	group_set_attrib_i(gdata, "Last", last);

	if (flags & QUEUE_ARTICLELIST_DOWNLOAD_NEW) {
		/* Add new headers to existing */
		if (gdata != Pan.active_group) {
			gdata->articlelist = db_load_articles(gdata);
		} else {
			articlelist_not_new(gdata->articlelist);
		}
		gdata->articlelist = g_slist_concat(gdata->articlelist, alist);

		db_save_articles(gdata, alist, count);	
		group_add_attrib_i(gdata, "Unread", count);
		group_add_attrib_i(gdata, "Total", count);
	} else {
		/* Save all the fresh new headers */
		articlelist_free(gdata->articlelist);
		gdata->articlelist = alist;
		db_save_articles(gdata, gdata->articlelist, count);
		group_set_attrib_i(gdata, "Unread", count);
		group_set_attrib_i(gdata, "Total", count);
	}

	return count;
}


/***************************************************************
* article_download 
***************************************************************/

int
nntp_article_download(group_data *gdata, article_data *adata)
{
	char *buffer;
	int count;
	FILE *f;
	
	gui_set_status_va(_("Downloading article '%s'"), adata->subject);

	if (set_group(gdata)) {
		network_error(_("Couldn't select group '%s'"), gdata->name);
		return -1;
	}

	if (put_server_va("HEAD %s\r\n", adata->number) == -1) {
		network_error(_("Error writing to socket."));
		return -1;
	}
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return -1;
	}
		
	if (strtol(buffer, (char **) NULL, 10) != 221) {
		return -1;
	}
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return -1;
	}
	
	f = acache_open_header(adata, ACACHE_NONE);
	while (strncmp(buffer, ".\r\n", 3)) {
		if (must_exit) {
			acache_close_header(f);
			acache_delete(adata->message_id);
			go_offline();
			return -1;
		}
		acache_putline_header(f, buffer);
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket."));
			acache_close_header(f);
			acache_delete(adata->message_id);
			return -1;
		}
	}
	acache_close_header(f);

	if (put_server_va("BODY %s\r\n", adata->number) == -1) {
		network_error(_("Error writing to socket."));
		return -1;
	}
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return -1;
	}
	if (strtol(buffer, (char **) NULL, 10) != 222) {
		return -1;
	}
	
	count = 0;
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return -1;
	}
	f = acache_open_body(adata, ACACHE_NONE);
	while (strncmp(buffer, ".\r\n", 3)) {
		if (must_exit) {
			acache_close_body(f);
			acache_delete(adata->message_id);
			go_offline();
			return -1;
		}
		acache_putline_body(f, buffer);
		gui_set_progress((float) count++ / (float) adata->linecount);
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket."));
			acache_close_body(f);
			acache_delete(adata->message_id);
			return -1;
		}
	}
	acache_close_body(f);
	gui_set_progress(1);

	return 0;
}

/***************************************************************
* article_decode 
***************************************************************/

int
nntp_article_decode(group_data *gdata, article_data *adata, GSList *list)
{
	char *buffer;
	float count = 0;
	GSList *l;
	FILE *f;
	article_data *adata0;

	/* Set the group */
	if (set_group(gdata)) {
		network_error(_("Couldn't select group '%s'"), gdata->name);
		return -1;
	} 

	/* Download bodies for all of the parts */
	for (l = list; l; l = l->next) {
		if ((adata0 = db_fetch_article(gdata, l->data)) == NULL) {
			gui_set_info_va(_("Couldn't find article header for message_id: %s"), l->data);
			article_add_flag(adata, STATE_ERROR);
			return -1;
		} 
		
		gui_set_status_va(_("Downloading part %d of %d: %s"), adata0->part, adata0->parts, adata0->subject);
		gui_set_progress(0);

		if (put_server_va("BODY %s\r\n", adata0->number) == -1) {
			network_error(_("Error writing to from socket."));
			article_free(adata0);
			return -1;
		}
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket."));
			article_free(adata0);
			return -1;
		}

		if (strtol(buffer, (char **) NULL, 10) != 222) {
			article_add_flag(adata0, STATE_ERROR);
			article_free(adata0);
			return -1;
		}
		count = 1;
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket."));
			article_free(adata0);
			return -1;
		}
		
		f = acache_open_body(adata0, ACACHE_KEEP);
		while (strncmp(buffer, ".\r\n", 3)) {
			if (must_exit) {
				acache_close_body(f);
				acache_delete(adata->message_id);
				article_free(adata0);
				return -1;
			}
			acache_putline_body(f, buffer);
			if ((buffer = get_server()) == NULL) {
				network_error(_("Error reading from socket."));
				acache_close_body(f);
				acache_delete(adata->message_id);
				article_free(adata0);
				return -1;
			}
			gui_set_progress(count++ / (float) adata0->linecount);
		}
		acache_close_body(f);
		article_free(adata0);
	}

	gui_set_progress(1);
	return 0;
}


void
nntp_send_noop (void)
{
	char *buffer;

	if (put_server("HELP\r\n") == -1) {
		network_error(_("Error writing to socket during HELP no-op."));
		return;
	}
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket during HELP no-op."));
		return;
	}
	if (atoi(buffer) == 100) {
 		while (strncmp(buffer, ".\r\n", 3)) {
			if ((buffer = get_server()) == NULL) {
				network_error(_("Error reading from socket during HELP no-op."));
				return;
			}
		}
	}
}
		
int connect_to_server(void)
{
	char *buffer;

	if (Pan.active_server == NULL) {
		pan_error_dialog(_("There is no active server.\nPlease configure one."));
		return 1;
	}
	gui_set_status_va(_("Connecting to server %s"), Pan.active_server->name);

	if (tcp_open(Pan.active_server, "nntp") < 0) {
		gui_clean();
		pan_error_dialog(_("Could not connect to server."));
		return 1;
	}
	if ((buffer = get_server()) == NULL) {
		gui_clean();
		pan_error_dialog(_("Could not get response from server."));
		return 1;
	}
	/* Now we attempt authentication.  This is not implemented 
	 * exactly as "Common NNTP Extensions" specifies, being that
	 * we didn't wait to be asked for it, but hopefully it 
	 * will work. */
	if ((strlen(Pan.active_server->username) + strlen(Pan.active_server->password)) > 1) {
		if (put_server_va("AUTHINFO user %s\r\n", Pan.active_server->username) == -1) {
			pan_error_dialog(_("Error writing to socket."));
			return 1;
		}
		if ((buffer = get_server()) == NULL) {
			pan_error_dialog(_("Error reading from socket."));
			return 1;
		}
		if (atoi(buffer) == 381) {
			/* 381 Means: More authentication information required */
			if (put_server_va("AUTHINFO pass %s\r\n", Pan.active_server->password) == -1) {
				pan_error_dialog(_("Error writing to socket."));
				return 1;
			}
			if ((buffer = get_server()) == NULL) {
				pan_error_dialog(_("Error reading from socket."));
				return 1;
			}
			if (atoi(buffer) != 281) {
				/* 281 Means: Authentication accepted */
				gui_clean();
				gui_set_info_va(_("Authentication unsuccessful."), buffer);
				pan_error_dialog(_("Authentication unsuccessful.\nServer said: %s"), buffer);
				return 1;
			}
		}
		/* Else authentication not required or not understood */
	}
	if (put_server("MODE READER\r\n") == -1) {
		pan_error_dialog(_("Error writing to socket."));
		return 1;
	}
	if ((buffer = get_server()) == NULL) {
		pan_error_dialog(_("Error reading from socket."));
		return 1;
	}
	if (atoi(buffer) == 200)
		Pan.active_server->posting = 1;
	else if (atoi(buffer) == 201)
		Pan.active_server->posting = 0;
	else {
		gui_clean();
		pan_error_dialog(_("Could not connect to server.\nServer said: %s"), buffer);
		return (1);
	}
	return (0);
}


/*
 * Menu callback functions 
 */
void go_offline_mf(void)
{
	if (is_online())
		must_exit = TRUE;
}

void stop_all_mf(void)
{
	if (is_online()) {
		must_exit = TRUE;
		stop_all = TRUE;
	}
}
